// This is the main DLL file.

#include <windows.h>
#include <stdio.h>
#include "out_bmp.h"

using namespace System::Windows::Forms;

namespace ImageSaver 
{
	bool ImageSaver::DumpErrorMessage(String^ msg, xml_node^ xn)
	{
		MessageBox::Show("BMP modul: " + msg + " fail");
		delete xn;
		return false;
	}

	String^ ImageSaver::GetExportModuleName()
	{
		return "BMP Bitmap";
	}

	int ImageSaver::GetAPIVersion()
	{
		return IMAGEFILTER_API_VERSION;
	}

	bool ImageSaver::Save_Image(String^ filename, IntPtr ptr, IntPtr size, IntPtr xml)
	{
		xml_node^ root = gcnew xml_node( xml.ToPointer() );
		xml_node^ alloc;
		int channelbits, colormapcount;
		int width, height, stride, packed, interleaved, channelcount, pixelsize;
		int *img_ptr = (int*)ptr.ToPointer();
		int *img_size = (int*)size.ToPointer();
		Byte* data;
		Byte* colormap;

		if (FilterSpace::ImageFilterBase::GetOneImageSetting(root, width, height, stride, packed, interleaved, channelcount, pixelsize, data))
			return DumpErrorMessage("GetOneImageSetting", root);

		if (channelcount != 3 && channelcount != 1) return DumpErrorMessage("ChannelCount", root);

		if (! root->XmlGetChildByName(alloc, "ChannelsBitAllocation")) return DumpErrorMessage("ChannelsBitAllocation", root);

		bool b = ((channelcount == 1) && (pixelsize == 1 || pixelsize == 4 || pixelsize == 8)) || (channelcount == 3);

		if (!b) return DumpErrorMessage("Channel reading", root);

		if (channelcount == 3)
		{
			colormapcount = 0;
			for (int i = 0; i < channelcount && b; ++i)
				if (b = alloc->XmlGetChildValueByName("Channel" + i.ToString(), channelbits)) 
					b = (channelbits == 8);
		}
		else if (channelcount == 1)
		{
			colormapcount = 1 << pixelsize;
			int colormapsize = colormapcount * sizeof(RGBQUAD);
			String^ SStr;
			bool    defaultcolormap = true;
			
			if ( alloc->XmlGetChildValueByName("ColorMap", SStr))
			{
				int length;
				base64(StringTostring(SStr), colormap, length);
				if (length < colormapsize)
					delete [] colormap;
				else
					defaultcolormap = false;
			}
			if (defaultcolormap)
			{
				//MessageBox::Show("Using default color map.");
				colormap = new Byte[colormapsize];
				for (int i = 0; i < colormapcount; i++)
				{
					colormap[i * 4 + 0] = colormap[i * 4 + 1] = colormap[i * 4 + 2] = (i * 255) / (colormapcount - 1);
					colormap[i * 4 + 3] = 0;
				}
			}
		}

		if (!b) return DumpErrorMessage("Channel reading", root);

		// --------------------------------------------
		// Framework guarantees, that stride is a dword aligned size

		BITMAPFILEHEADER fileheader;
		fileheader.bfType      = *(unsigned short int*)("BM");
		fileheader.bfSize      = sizeof(BITMAPFILEHEADER) + sizeof(BITMAPINFOHEADER) + sizeof(RGBQUAD) * colormapcount + stride * height;	
		fileheader.bfReserved1 = 0;
		fileheader.bfReserved2 = 0;
		fileheader.bfOffBits   = 54 + sizeof(RGBQUAD) * colormapcount;

		BITMAPINFOHEADER header;
  		header.biSize          = sizeof(BITMAPINFOHEADER);
		header.biWidth         = width;
		header.biHeight        = height;
		header.biPlanes        = 1;
		header.biBitCount      = pixelsize;
		header.biCompression   = BI_RGB;
		header.biSizeImage     = 0;
		header.biXPelsPerMeter = 0;
		header.biYPelsPerMeter = 0;
		header.biClrUsed       = 0;
		header.biClrImportant  = 0;

		char* fileName = StringToCharstring(filename);
		FILE *fp;
		errno_t err_msg = fopen_s(&fp, fileName, "wb");
		delete [] fileName;

		if (err_msg != 0) return DumpErrorMessage("open file: " + filename, root);

		fwrite((BYTE *)&fileheader, sizeof(BITMAPFILEHEADER), 1, fp);
		fwrite((BYTE *)&header    , sizeof(BITMAPINFOHEADER), 1, fp);

		if (colormapcount > 0)
		{
			fwrite(colormap, 1, sizeof(RGBQUAD) * colormapcount, fp);
			delete [] colormap;
		}

		*img_size = height;

		long rowidx, row_size = stride;

		for ( long row = 0 ; row < height ; ++row ) 
		{
			rowidx = (long unsigned)(height - row - 1) * stride;						      
			if ( fwrite( (void*)(data + rowidx), stride, 1, fp ) == -1 ) break;
			(*img_ptr)++;
		}

		fclose(fp);

		// --------------------------------------------

		delete root;
		return true;
	}
}